@@ -1,18 +1,23 @@
 """
 Libreria per fare grafici di fit e tabelle LaTeX a partire da dati raccolti in un file Excel
-TODO: graph log
-TODO: fit ricorsivo per errori sulla x con ciclo che termina quando non cambia più la soluzione oppure usare algoritmi esistenti che includono errore su x
-
+TODO: graph log con p value
+TODO: rivedere esempi documentazione
 """
 __docformat__="numpy"
 
+from curses import use_env
+from email import header
+from ntpath import join
 import numpy as np
 from matplotlib import pyplot as plt
 from scipy.optimize import curve_fit
 from scipy.odr import *
+from scipy import stats
 import pandas as pd
 from pint import UnitRegistry
-import re 
+import re
+
+from sympy import true 
 
 def _tex_escape(text):
     """
@@ -46,20 +51,44 @@ def _converti_latex(text):
     new_text=new_text.replace("[]",r"\(\left["+um_tex+r"\right]\)") #inserimento siunix dopo aver fatto l'escape dei caratteri speciali per latex
     return new_text
 
-def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_latex,
+def _deriva(funzione, x0, parametri, precisione=0.00000001):
+    derivata_x0=[]
+    for element in x0:
+        old=0
+        derivata=1
+        #TODO: secondo me funziona strano ci sono sicuramente errori numerici e non so se trascurabili
+        #limite numericamente calcolato
+        while np.abs(old-derivata)>precisione:
+            old=derivata
+            n=-10
+            epsilon=np.power(10.0, n)
+            derivata=(funzione(element+epsilon, *parametri) - funzione(element, *parametri))/epsilon
+            n-=10
+        derivata_x0.append(derivata)
+
+    return derivata_x0
+
+
+def fit(excel_name,file_result,f_fit,popt_names_latex,variabile_x,variabile_y,sigma_y,
     sigma_x=None,
+    use_ODR=False,
+    numero_iterazioni=10,
     split_graphs=False,
     titoli=["",""],
     xlabel=None,
     ylabel=None,
     chi=True,
+    ridotto=True,
     x_chi=0,
     y_chi=0,
     post_decimal_chi=0,
     log_to_file=False,
+    nolatex=False,
     path=".",
     xscale="linear",
     yscale="linear",
+    lim_inf=None,
+    lim_sup=None,
     griglia_fit={
         "which" : "both",
         "ls" : "dashed",
@@ -75,9 +104,12 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
     style_residui={"fmt" : "."},
     res_line_color="navy",
     pad_percent=0.10,
+    precisione_curvatura=1001,
     foglio="Foglio1",
     funzioni_modifica=[None, None],
-    fig_dimension=(6.24136,4.4)
+    residui_ratio=[4, 1],
+    fig_dimension=(6.24136,4.4),
+    estensione_grafico=".pdf"
 ):
     """
     funzione che genera:
@@ -92,28 +124,32 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
     
     Parameters
     -------------------------
-    file_name : str
+    excel_name : str
         nome file excel da elaborare
 
     file_result : str
         nome da dare ai pdf generati
 
-    variable_x, variable_y, sigma_y : str
-        ogni parametro contine il valore dell'header della colonna del foglio excel da cui prendere i dati
-        (deve contenere l'unità di misura nel formato: "[unità di misura]")
-
     f_fit : func
         funzione che rappresenta la curva g(x) utilizzata nel fit.
         il primo parametro di f è x, gli altri sono tutti i parametri che definiscono g(x)
-
-    titoli_p_latex : list
+    
+    popt_names_latex : list
         elenco dei nomi dei parametri ottimali che vengono ricavati dal fit
         verranno usati come contenuto della cella del titolo di una tabella latex contenente i valori numerici dei parametri
+        devono essere scritti in Latex
+
+    variable_x, variable_y, sigma_y : str
+        ogni parametro contine il valore dell'header della colonna del foglio excel da cui prendere i dati
+        (deve contenere l'unità di misura nel formato: "[unità di misura]")
     
     sigma_x: str, optional
         contine il valore dell'header della colonna del foglio excel da cui prendere i dati
         (deve contenere l'unità di misura nel formato: "[unità di misura]") (default: None)
-        se un valore viene specificato viene usato un metodo di fit con Orthogonal Distance Regression, altrimenti si usa scypi.curvefit()
+        se un valore diverso d "None" viene specificato, viene usato un metodo di fit iterato con sigma efficaci, altrimenti si usa scypi.curvefit()
+
+    use_ODR: bool, optional
+        se True viene eseguito il fit con Orthogonal Distance Regression invece che iterando un fit del minimo chi^2 con sigma efficaci
 
     split_graphs: bool, optional
         se True la funzione genera due file separati per il grafico dei residui e il grafico di fit (default: False)
@@ -127,7 +163,12 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
         del file excel verranno utilizzati, convertendoli automaticamente in LaTeX ed interpretando um tra parentesi quadre (default: None)
 
     chi : bool, optional
-        se True allora sul grafico viene mostrato il valore del chi quadro e il numero di gradi di libertà (default: True)
+        se True allora sul grafico viene mostrato il valore del chi quadro (default: True)
+
+    ridotto : bool, optional
+        se True allora sul grafico viene mostrato il valore del chi quadro ridotto,
+        se False viene mostrato il chi quadro e i gradi di linertà (default: True)
+
 
     x_chi, y_chi : int, optional
         coordinate del punto in cui scrivere il valore del chi quadro e il numero di gradi di libertà (default: 0, 0 --> primo punto del plot)
@@ -138,12 +179,18 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
     log_to_file : bool, optional
         if True le informazioni sulla bontà del grafico sono messe in un file di testo. If False le informazioni possono essere lette in console. (default: False)
 
+    nolatex : bool, optional
+        if True il grafico è disegnato senza compilare in latex
+
     path : str, optional
         path directory di partenza (default: ".")
 
     xscale, yscale : str, optional
         tipo di scala da utilizzare per l'asse corrispondente del grafico (default: 'linear')
 
+    lim_inf, lim_sup : float, optional
+        valori minimo e massimo rappresentati sull'asse delle x nella creaione del grafico. Se il valore è None il grafico si adatta al set di dati (default: None)
+
     griglia_fit, griglia_residui : dict, optional
         dizionario contenente i parametri opzionali da passare a pyplot.grid() (default: {"which" : "both", "ls" : "dashed", "color" : "powderblue"})
 
@@ -153,9 +200,12 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
     fit_line_color, res_line_color : str, optional
         colori da applicare alla curva di fit e alla linea dello zero nel grafico dei residui (default: "darkorange", "navy")
 
-    pad_pecent: float, optional
+    pad_pecent : float, optional
         dimensione percentuale dello spazio del grafico in più rispetto al minimo necessario a contenere i dati (default: 0.10)
 
+    precisione_curvatura : int, optional
+        numero di punti utilizzato per il disegno della curva (default: 1001)
+
     foglio : str, optional
         nome del foglio da cui prendere i dati (default: "Foglio1")
 
@@ -171,9 +221,15 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
             
         - data: panda dataframe rappresentante il foglio Excel
 
+    residui_ratio : [int, int], optional
+        rapporto dimensione (altezza), tra grafico e grafico dei residui (default: [4,1])
+
     fig_dimensions : (float, float), optional
         tupla contenete le dimensioni (x,y) dell'immagine creata in inches (default: (6.24136,4.5))
 
+    estensione_grafico : str, optional
+        estensione in cui viene salvato il grafico (default: ".pdf")
+
     Examples
     ----------------------------
     >>> def retta(x,m,q):
@@ -217,8 +273,8 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
         if result:
             print(result)
     """
+ 
     #definizione di una funzione della forma utile all'utilizzo di ODR
-    #TODO: vedere se è veramente necessario
     def f_fit_odr (Beta, variabile):
         return f_fit(variabile, *Beta)
 
@@ -234,93 +290,130 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
     if sigma_x:
         if "[" not in sigma_x or "]" not in sigma_x:
             return "Manca unità di misura in dx"    
-    if not file_name:
+    if not excel_name:
         return "specificare nome file excel"
     if not file_result:
         return "specificare nome file pdf risultato (senza .pdf)"
+    if not popt_names_latex:
+        return "specificare i nomi dei parametri ottimmali di fit (scritti in latex)"
 
     #lettura file
-    excel = pd.read_excel(path+'\\'+file_name, sheet_name=foglio)
+    excel = pd.read_excel(path+'\\'+excel_name, sheet_name=foglio)
 
-    t=[e for e in excel[variabile_x].tolist() if e == e] #NaN != NaN quindi il test impedisce di scrivere i valori nulli
-    y=[e for e in excel[variabile_y].tolist() if e == e]
+    t=[e for e in excel[variabile_x].tolist() if e == e] #t perché x non si poteva usare per  ambiguità
+    y=[e for e in excel[variabile_y].tolist() if e == e] #NaN != NaN quindi il test impedisce di scrivere i valori nulli
     dy=[e for e in excel[sigma_y].tolist() if e == e]
     if sigma_x:
         dx=[e for e in excel[sigma_x].tolist() if e == e]
 
     #fit
-    popt, pcov = curve_fit(f_fit,t,y,sigma=dy)
 
-    if sigma_x:
+    #fit iniziale
+    popt, pcov = curve_fit(f_fit,t,y,sigma=dy,absolute_sigma=true)
+    #sigma efficace
+    if dx:
+        sigma_eff=np.sqrt(np.power(dy,2)+np.power(_deriva(f_fit, t, popt)*dx,2)) #potrebbe non funzionare
+    else:
+        sigma_eff=dy
+        dx=np.zeros(len(t))
+    
+    #fit iterato
+    for i in range(numero_iterazioni):
+        sigma_eff=np.sqrt(np.power(dy,2)+np.power(_deriva(f_fit, t, popt)*dx,2)) #potrebbe non funzionare
+        popt, pcov = curve_fit(f_fit,t,y,sigma=sigma_eff,absolute_sigma=true)
 
+    #ODR
+    if use_ODR:
         linear_model = Model(f_fit_odr)
         data = RealData(t,y,dx,dy)
         odr = ODR(data,linear_model,beta0=popt)
         output = odr.run()
         popt = output.beta
+        pcov = output.cov_beta
 
-    chi=sum(np.power((y-f_fit(np.array(t),*popt)), 2)/np.power(dy,2))
+
+
+    #chi^2
+    
+    if not use_ODR:
+        chisq=sum(np.power((y-f_fit(np.array(t),*popt)), 2)/np.power(sigma_eff,2))
+    else:
+        chisq=output.sum_square
     gradi_di_liberta=len(t)-len(popt)
-    chi_rid=chi/gradi_di_liberta    
-
+    chi_rid=chisq/gradi_di_liberta    
+ 
     #valutazione grafico
     sopra_fit=0
     sotto_fit=0
     lontano_fit=0
-    for element_y, element_t, element_dy in zip(y, t, dy):
+    for element_y, element_t, element_dy in zip(y, t, sigma_eff):
         if element_y-f_fit(element_t,*popt)>0:
             sopra_fit+=1
         if element_y-f_fit(element_t,*popt)<0:
             sotto_fit+=1
         if np.absolute(element_y-f_fit(element_t,*popt))>element_dy:
             lontano_fit+=1
+    p_value=stats.distributions.chi2.sf(chi_rid, 1)
 
     frase_log=[]
     frase_log.append("Numero di punti sopra alla funzione di best fit: "+str(sopra_fit)+"\n"+"Numero di punti sotto alla funzione di best fit: "+str(sotto_fit))
     frase_log.append("(expected: " + str(len(t)/2)+")")
-    frase_log.append("Numero di punti che distano dal grafico n>1 barre di errore: "+str(lontano_fit)+" (expected: "+str(0.32 *len(t))+")")
+    frase_log.append("Numero di punti che distano dal grafico n>1 barre di errore: "+str(lontano_fit)+" (expected-ipotesi  errori gaussiani: "+str(0.32 *len(t))+")")
+    frase_log.append("la probabilità di ottenere un valore più estremo è di: "+p_value)
     if log_to_file:
         with open(path+"\\"+file_result+"_fit_log.txt", 'w') as fit_log:
             fit_log.write("\n".join(frase_log))
             fit_log.close()
     else:
-        print(frase_log[0])
-        print(frase_log[1])
-        print(frase_log[2])
+        for e in frase_log:
+            print(e)
 
-    #scrittura file_tabella.tex
-    empty_space="&"*(len(popt)-1)+r"\\" #linee vuote per aggiungere spazio prima e dopo in ogni riga
-    lines_prima=[r"\renewcommand{\arraystretch}{0.5}",r"\begin{table}[h!]",r"\centering"]
-    colonne="c"*len(popt) #colonne centrate in numero variaabile a seconda di len(popt)
-    lines_prima.append(r"\begin{tabular}{"+colonne+r"}")
-    lines_prima.append(r"\toprule")
-    lines_prima.append(empty_space)
-    lines_prima.append(r" & ".join(titoli_p_latex)+r" \\")#titoli
-    lines_prima.append(empty_space)
-    lines_prima.append(r"\midrule")
-    lines_prima.append(empty_space)
+
+    #scrittura file: *_tabella.tex con i risultati di fit (longtable)
+
+    colonne="c"*len(popt) # colonne centrate in numero pari al numero di parametri
+    line_begin=r"\begin{longtable}{"+colonne+r"}"
+    table_header="&".join(popt_names_latex)+r"\\"
+    longtable_header=[r"\toprule",table_header ,r"\midrule", r"\endfirsthead", r"\toprule",table_header ,r"\midrule",r"\endhead", r"midrule", r"\multicolumn{3}{r}{{continua a pagina seguente}} \\", r"\midrule", r"\endfoot", r"\bottomrule", r"\endlastfoot"]
     string_popt=[str(i_opt) for i_opt in popt] #numeri convertiti in stringa
-    lines_prima.append(r" & ".join(string_popt)+r" \\") #valori
-    lines_prima.append(empty_space)
-    lines_dopo=[r"\bottomrule",r"\end{tabular}",r"\caption{risultati del fit}",r"\end{table}",r"\renewcommand{\arraystretch}{1}"]
-    lines_prima.extend(lines_dopo)    
+    data_table=r" & ".join(string_popt)+r" \\"
+    line_end=r"\end{longtable}"
+
+    tabella=[] #creazione tabella completa
+    tabella.append(line_begin)
+    tabella.extend(longtable_header)
+    tabella.append(data_table)
+    tabella.append(line_end)
     
-    with open(path+"\\"+file_result+"_tabella.tex", 'w') as tabella_tex:
-        tabella_tex.write("\n".join(lines_prima))
+    with open(path+"\\"+file_result+"_tabella.tex", 'w') as tabella_tex: #scrittura file .tex
+        tabella_tex.write("\n".join(tabella))
         tabella_tex.close()
 
     #disegno grafico 
-    plt.rcParams["figure.figsize"] = fig_dimension #dimensioni
+    plt.rcParams["figure.figsize"] = fig_dimension #dimensioni impostate come richiesto (utile per corretta dimensione dei caratteri)
 
-    pad=(max(t)-min(t))*pad_percent #aggiunta spazio
-    lim_inf=min(t)-pad if min(t)-pad>0 else 0
-    lim_sup=max(t)+pad
+    #aggiunta di spazio prima e dopo il range di dati rappresentato
+    pad=(max(t)-min(t))*pad_percent
+    if not lim_inf:
+        if min(t)-pad>0 or min(t)<0:
+            lim_inf=min(t)-pad
+        else: 
+            lim_inf=0
+    if not lim_sup:
+        if max(t)+pad<0 or max(t)>0:
+            lim_sup=max(t)+pad
+        else: 
+            lim_sup= 0
 
-    a1 = np.linspace(lim_inf, min(t), 1001) #separazione per tratteggio
-    a2 = np.linspace(min(t), max(t), 1001)
-    a3 = np.linspace(max(t), lim_sup, 1001)
 
-    plt.rcParams.update({"text.usetex" : True,   #LaTeX
+    #categorizzazione punti reali e punti previsti dal modello con tratteggio e linea continua
+    a1 = np.linspace(lim_inf, min(t), precisione_curvatura) #tratteggio prima
+    a2 = np.linspace(min(t), max(t), precisione_curvatura) #continua
+    a3 = np.linspace(max(t), lim_sup, precisione_curvatura) #tratteggio poi
+
+    #TODO: nolatex mode for faster debug
+    #Latex parameters
+    plt.rcParams.update({"text.usetex" : True,   #utilizza latex nella compilazione del grafico
         "font.family": "computer modern",
         "text.latex.preamble": "\n".join([ # plots will use this preamble
         r"\usepackage[utf8]{inputenc}",
@@ -328,23 +421,20 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
         r"\usepackage{siunitx}",
         r"\usepackage{amsmath}",
         ])})
-    
+
+    #default label impostati a nome della ccolonna di excel usata per estrarre i dati   
     if not xlabel:
         xlabel=_converti_latex(variabile_x)
 
     if not ylabel:
         ylabel=_converti_latex(variabile_y)
 
-    if x_chi==0:
-        x_chi=t[0]*1.1
-    if y_chi==0:
-        y_chi=y[0]*1.1
-
+    #creazione oggetti plt usando sempre metodi subplot
     if split_graphs:
         fig1, ax1 = plt.subplots(1, 1)
         fig2, ax2 = plt.subplots(1, 1)
     else:
-        fig1, (ax1, ax2) = plt.subplots(2, 1, gridspec_kw={'height_ratios': [4, 1]}, sharex=True)
+        fig1, (ax1, ax2) = plt.subplots(2, 1, gridspec_kw={'height_ratios': residui_ratio}, sharex=True)
 
     #grafico di fit
     ##aspetto
@@ -352,9 +442,9 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
     ax1.set_xscale(xscale)
     ax1.set_yscale(yscale)
     ax1.grid(**griglia_fit)
-    ##plot
-    ax1.errorbar(t, y, dy, **style_fit)
-    ax1.plot(a1, f_fit(a1, *popt), linestyle="dashed", color=fit_line_color) # l'asterisco indica che popt viene espanso passando i parametri singolarmente e non come lista.
+    ##plot TODO: verificare che funzioni con dx=0
+    ax1.errorbar(t, y, dy, dx, **style_fit)
+    ax1.plot(a1, f_fit(a1, *popt), linestyle="dashed", color=fit_line_color)
     ax1.plot(a2, f_fit(a2, *popt), color=fit_line_color)
     ax1.plot(a3, f_fit(a3, *popt), linestyle="dashed", color=fit_line_color)
     ##adjust limits
@@ -363,29 +453,40 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
     #residui
     ##aspetto
     ax2.set_title(titoli[1])
-    ax2.margins(0, 0.2)
+    ax2.margins(0, 0.2) #TODO: controllare perché ho fatto questa cosa
     ax2.set_xscale(xscale)
     ax2.set_yscale(yscale)
     ax2.grid(**griglia_residui)
     ##plot
-    y_res=y-f_fit(np.array(t), *popt)  # l'asterisco indica che popt viene espanso passando i parametri singolarmente e non come lista.
+    y_res=y-f_fit(np.array(t), *popt)
     ax2.errorbar(t, y_res, dy, **style_residui)
-    ax2.plot(a1, a1-a1, color=res_line_color)
+    ax2.plot(a1, a1-a1, color=res_line_color) #evidenzia la linea dello zero
     ax2.plot(a2, a2-a2, color=res_line_color)
     ax2.plot(a3, a3-a3, color=res_line_color)
-    ##adjust for x axi symmetry
+    ##adjust ylim for x axis symmetry
     low, high = ax2.get_ylim() 
     bound = max(abs(low), abs(high)) # find the new limits
     ax2.set_ylim(-bound, bound) # set new limits
-    ax2.set_xlim(min(a1),max(a3))  
+    ax2.set_xlim(min(a1),max(a3))  # manual share x (così funziona anche per split_graph=True )
 
     #chi^2
+    #TODO: posizionamento del chi^2
+    if x_chi==0:
+        x_chi=t[0]
+    if y_chi==0:
+        y_chi=y[0]+0.1*(max(y)-min(y))
+
     if chi:
-        ax1.text(x_chi,y_chi,r"$$\frac{\chi^2}{\nu}="+str(round(chi_rid,post_decimal_chi))+r"$$")
+        if ridotto:
+            ax1.text(x_chi,y_chi,r"$$\frac{\chi^2}{\nu}="+str(round(chi_rid,post_decimal_chi))+r"$$")
+        else:
+            ax1.text(x_chi,y_chi,r"$$\chi^2 / \nu}="+str(round(chi,post_decimal_chi))+r" / "+gradi_di_liberta+r"$$")
+            
 
     fig1.supxlabel(xlabel, y=0.05)
     fig1.supylabel(ylabel, x=0.03)
 
+    #possibilità di modificare un po' tutto con una funzione arbitraria per disegnare cose extra
     if funzioni_modifica[0]:
         funzioni_modifica[0](fig1, ax1,t,y,popt, excel)  
 
@@ -400,17 +501,17 @@ def fit(file_name,file_result,variabile_x,variabile_y,sigma_y,f_fit,titoli_p_lat
         fig2.tight_layout()
         fig2.canvas.manager.set_window_title("Residui") 
 
-        fig2.savefig(path+'\\'+file_result+'_residui.pdf')
+        fig2.savefig(path+'\\'+file_result+'_residui'+estensione_grafico)
     else:
         if funzioni_modifica[1]:
             funzioni_modifica[1](fig1, ax2,t,y,popt, excel)     
 
     fig1.tight_layout()
     fig1.canvas.manager.set_window_title("Grafico") 
-    fig1.savefig(path+'\\'+file_result+'.pdf')
+    fig1.savefig(path+'\\'+file_result+estensione_grafico)
     plt.show()
 
-def tabella(file_name,file_result,
+def tabella(excel_name,file_result,
     path=".",
     colonne=None,
     foglio="Foglio1",
@@ -426,7 +527,7 @@ def tabella(file_name,file_result,
 
     Parameters
     -------------------------
-    file_name : str
+    excel_name : str
         nome file excel da elaborare
 
     file_result : str
@@ -460,18 +561,18 @@ def tabella(file_name,file_result,
     >>> my_path="C:\\Users\\cremo\\Documents\\Università\\Relazioni\\pendolo_json"
         file_result="nome_grafico"
         file_tabella="nome_tabella"
-        file_name="prova_json.xlsx"
+        excel_name="prova_json.xlsx"
         formati={
             "tempi_medi [s]": "{:.4f}".format,
             "sigma_t [s]": "{:.9f}".format,
             "lunghezze [m]": "{:.3f}".format
         }
-        result=js.tabella(file_name,file_tabella, formatter=formati, path=my_path)
+        result=js.tabella(excel_name,file_tabella, formatter=formati, path=my_path)
         if result:
             print(result)
     """
     #lettura file
-    excel = pd.read_excel(path+'\\'+file_name, sheet_name=foglio, usecols=colonne)
+    excel = pd.read_excel(path+'\\'+excel_name, sheet_name=foglio, usecols=colonne)
     #scrittura file
     ##creazione header in stile LaTeX
     my_header=[]
